; dat     segment

; WARNING: nasm treats mov ax,label and mov ax,[label] the same,
; use mov ax,[seg:label] to access memory AT label !?
; (only a problem for sources, not for destinations)

        org  0	; SYS/XMS version of an hard disk read cache
		; inspired by PC Magazine dcache.com (which was
		; COM/EMS version and only 1 drive, max 256 MBy drive)
	; HDDCACHE written by Eric Auer 10/2001 eric@coli.uni-sb.de
        ; (based on 5/96 sys with 386 driver for 850 MB HDD also by EA)

; not implemented: floppy cache, CD-ROM cache, write cache, cmd line args,
; some verbosity, int 0x13 LBA extensions support (64bit LBA impressive...)
; not planned: useability with EMS and 8086, DOS rather than int 0x13 cache
; history: first working version (fixed 1024 sec buffer) 12./13.10.2001
;          added more verbosity, statistics, round-robin search 14.10.2001

;        assume  cs:dat,ds:dat,es:dat

next    dw -1,-1
attrib  dw 1000000000000100b	; very simple NUL device
stra    dw strat	; strategy  -> prepare, save ES:BX pointer
inr     dw intr		; interrupt -> main procedure
nam     db 'HDCACHE$'	; our name, WARNING: becomes invalid file
			; name, this is why the "$" is included
			; ("HDCACHE$.*" being invalid is less annoying)

pb      dd 0		; buffer for the ES:BX pointer...

sectors	dw 1024		; how many sectors to cache
tabpos	dw table	; where to start the search (round-robin)
xmshandle
	dw 0		; our XMS handle
xmsvec	dd 0		; call far pointer for XMS services
geometry
	db 63,255	; for the hard disks 0x80 .. 0x83:
	db 63,255	;   first byte is max sector number,
	db 63,255	;   second byte is max head number
	db 63,255
			; next is the XMS move command param structure:
xmsinf	dd 512	; 0 size to move
	dw 0	; 4 source handle, 0 for dos
	dd 0	; 6 source offset (or pointer for dos)
	dw 0	; A destination handle...
	dd 0	; C destination offset...

meepwhy	dw 0	; pointer to last meep" error message
running	db 0		; 1 means "init already done"
diskinf	db 0	; last disk access status
	; *** meepwhy and diskinf <-> reentrancy, see below...
	; *** (no real problem, only diagnostics may be wrong)

	db 'INF>'
rdhit	dd 0	; those 4 counters offer some statistics
rdmiss	dd 0
wrhit	dd 0
wrmiss	dd 0
	db '<INF'


.8086
strat:	; in theory, we could even copy the data at es:bx here
	mov word [cs:pb+2],es
	mov word [cs:pb],bx
	RETF


intr:		
	pushf
        push bx
        push es
        les bx,[cs:pb]
        mov byte [es:bx+0x17],0		; no error
        and word [es:bx+3],0x7ffd	; not busy / no error
        or word [es:bx+3],1		; ack status
        cmp byte [es:bx+2],0	; is the command INIT ?
	jnz nix
	test byte [cs:running],1
	jnz nix				; only allow init once
        jmp install
nix:    pop es
        pop bx
        popf
        RETF

; -------------------------------------------------------------------

	; the new INT 0x13 handler follows (only for 80386+XMS):
.386
NEWI13:	or dl,dl	; only cache hard drives for now
bailout:
	jns short oldint	; else, skip completely to old handler
	cmp dl,0x83	; only cache first 4 drives (2 bits...)
	ja short oldint
	cmp ah,5        ; block format requests (5/6/7)
	jb rwmaybe	; 0..4 are most normal
	cmp ah,8	; another format thing
	jb short enderr
	jz short oldint
	cmp ah,9
	jz short fishyfishy	; REALLY bail out if geometry is changed!
	cmp ah,0x0b
	jz short fishy	; flush cache if somebody writes long (with ECC)!
;	cmp ah,0x18		; set media type for format: better bail out?
;	jz short fishyfishy	; func 0x18 is also used for harmless stuff
	cmp ah,0x22
	jz short fishyfishy	; PS/2 write incompatible to NEXTSEC
	cmp ah,0x21
	jz short fishyfishy	; PS/2 read incompatible to NEXTSEC
	cmp ah,0x43
	jz short fishy	; flush cache if 64bit LBA write occurs (int 13 ext)
	cmp ah,0x46
	jz short fishy	; flush cache if media is ejected (int 13 ext)
	cmp ah,0x4a
	jz short fishyfishy	; STOP cache if CD emulates boot drive
	cmp ah,0x4c
	jz short fishyfishy	; STOP cache if CD emulates boot drive
	cmp ax,0x50
	jae short fishyfishy	; stop cache completely any undocumented stuff
	jmp short oldint	; other mixed stuff is not even worth flushing
	; *** this "dangerous function enumeration" is potentially dangerous

	;   a (read long), c (seek), d (recalib+reset), 10 (ready?), 
	;   11 (recalib), 12 (ram diag), 13 (disk diag), 14 (ctrl diag)
	; e/f are r/w XT sector buffer, used for diagnostic and format...
	; 19: park - would be a good time to flush if we had a write cache
	; 1a: esdi format, but also harmless stuff. 1b...: mixed
	; 1f: syquest lock door... 20...: atapi stuff and mixed, very mixed!
	; 21/22: read/write multiple (ps/2): dh lsb are 2 more cyl bits!

	; 41... are the next REAL (generic) functions: int 13 ext
	; 42 - reading with 64bit LBA - is kind of harmless for us, ignored!
	; (43 write, 44 verify, 45 lock/unlock, 46 eject, 47 seek, 48 info,
	;  49 changed?, 4a..4d bootable cdrom, 4e set dma/pio)
	; above 4f is again very mixed

fishyfishy:	; UNINSTALL because resetting the drive params may bend geom
	mov byte [cs:bailout],0xeb	; PATCH: make skipping unconditional
	mov word [cs:meepwhy],fish2err
		call meep	; complain a bit...
	jmp short oldint
fishy:
		call flush
	mov word [cs:meepwhy],fisherr
		call meep	; complain again!
	jmp short oldint
rwmaybe:
			; add call meep here if you are REALLY curious
	cmp ah,3
	jz hdwrite	; copy to cache if write
	cmp ah,2
	jz hdread	; read, maybe from cache
	jmp short oldint	; other 0..4 are safe
oldint:	db 0eah		; JMP DWORD ...
oldi13a:	dd 0f000a25fh   ; old int 0x13 vector

enderr:	mov ah,4
	stc
	RETF +2
endok:	clc		; ah has to contain error value...
	RETF +2

; -------------------------------------------------------------------

	; es:bx is buffer, cx and dx are location, al is size

hdwrite:
	pusha		; loop for all sectors: copy to cache
	push ax		; some samity check ...
	mov ah,al
	mov al,0
	add ax,ax	; ... ax becomes 0x200 * al ...
	add ax,bx
	pop ax
	jc short dmafail	; ... big stink if 64k boundary crossed
	or al,al
	jz short wrdone
		; find or create new destination bin AX, write data there,
		; and (for now) write all sectors at once after all writing
		; to cache is done. Flush whole cache on write error for now.
		; *** Does NOT ensure perfect sync that way!
wrsec:	push ax
	inc dword [cs:wrhit]	; statistics
		call findbin	; find existing bin, if any -> AX
	jnc short wrnotnew
	dec dword [cs:wrhit]	; statistics
	inc dword [cs:wrmiss]	; statistics
		call newbin	; create a new bin otherwise -> AX
wrnotnew:
		call copytoxms
	pop ax
	cmp al,1
	jz short wrdone	; was the last...
		call nextsec
	dec al
	jmp short wrsec
wrdone:
	popa	
		call callold	; do the real writing
	jnc endok		; everything ok, cache still valid
	mov word [cs:meepwhy],wrerr
		call meep	; warn user
		call flush	; *** panic and flush cache on write error
	stc
	RETF +2

dmafail:
	popa
	mov ah,9	; 64k boundary error
	stc
	RETF +2

; -------------------------------------------------------------------

hdread:
	pusha		; loop for all sectors: read from cache or disk
	push ax		; some samity check ...
	mov ah,al
	mov al,0
	add ax,ax	; ... ax becomes 0x200 * al ...
	add ax,bx
	pop ax
	jc short dmafail	; ... big stink if 64k boundary crossed
	or al,al
	jz short rddone
		; find source bin AX, read from XMS if found.
		; IF none found, read from disk (panic if error),
		; create new bin, save a copy of the read data in XMS
rdsec:	push ax
		call findbin	; find sector in XMS -> bin AX
	jc notcached
	inc dword [cs:rdhit]	; statistics
		call copytodos	; if found, read data from XMS bin AX
	jmp short rdloop
notcached:
	inc dword [cs:rdmiss]	; statistics
	push es
	push bx
	push cx
	push dx
	mov ax,0x0201	; read only one sector
		call callold	; really read data from disk
	pop dx
	pop cx
	pop bx
	pop es
	mov [cs:diskinf],ah	; save error code (see warning below)
	jc short rdfail	; *** could be better than just bailing out!
		call newbin	; find a space in XMS -> bin AX
		call copytoxms	; copy data to XMS
rdloop:
	pop ax
	cmp al,1
	jz short rddone	; was the last...
		call nextsec
	dec al
	jmp short rdsec
rddone:
	popa
	mov ah,0	; return simple success (who needs ECC warnings ;-))
	clc
	RETF +2
rdfail:
	pop ax
	popa
	mov ah,[cs:diskinf]
	mov word [cs:meepwhy],rderr
		call meep	; give feedback...
	or ah,ah
	jz rdfsane
	mov ah,4	; read error
rdfsane:		; *** the sanity check above is because using
	stc		; *** diskinf rather than stack is not reentrant
	RETF +2

; -------------------------------------------------------------------

	; XMS helper functions

copytoxms:	; copy 1 sector from es:bx to xms bin AX
	pushf
	push ds
	push cs
	pop ds
	pusha	; do not trust XMS
	push eax
	mov si,xmsinf
	cli			; *** even better: create xmsinf on stack
	mov dword [ds:si],512	; size of sector: 512 byte
	mov word [ds:si+0x04],0	; source handle: dos
	push word [ds:xmshandle]
	pop word [ds:si+0x0a]	; destination: xms
	mov [ds:si+0x06],bx	; source pointer offset
	mov [ds:si+0x08],es	; source pointer segment
	movzx eax,ax		; clear high 16 bits
	shl eax,9		; assume 512 byte sectors!
	mov [ds:si+0x0c],eax	; destination offset in xms
	mov ah,0x0b	; XMS copy command
		call far [cs:xmsvec]
	or ax,ax
	jnz wrxmsok
	; *** whoops: XMS copy returned an error in BL, BL trashed !
		mov al,bl	; for debugging
		mov ah,0x0b	; same
	mov word [cs:meepwhy],xmserr
		call meep
wrxmsok:
	pop eax
	popa
	pop ds
	popf
	ret


copytodos:	; copy 1 sector from xms bin AX to es:bx
	pushf
	push ds
	push cs
	pop ds
	pusha	; do not trust XMS
	push eax
	mov si,xmsinf
	cli			; *** even better: create xmsinf on stack
	mov dword [ds:si],512	; size of sector: 512 byte
	push word [ds:xmshandle]
	pop word [ds:si+0x04]	; source handle: xms
	mov word [ds:si+0x0a],0	; destination: dos
	movzx eax,ax		; clear high 16 bits
	shl eax,9		; assume 512 byte sectors!
	mov [ds:si+0x06],eax	; source offset in xms
	mov [ds:si+0x0c],bx	; destination pointer offset
	mov [ds:si+0x0e],es	; destination pointer segment
	mov ah,0x0b	; XMS copy command
		call far [cs:xmsvec]
	or ax,ax
	jnz rdxmsok
	; *** whoops: XMS copy returned an error in BL, BL trashed !
		mov al,bl	; for debugging
		mov ah,0x0b	; same
	mov word [cs:meepwhy],xmserr
		call meep
rdxmsok:
	pop eax
	popa	; <--
	pop ds
	popf
	ret

; -------------------------------------------------------------------

	; helper function: increment CHS sector number (uses stored geometry)

nextsec:
	add bx,0x200	; 512by sectors: advance in buffer
	push bx
	push ax		; go to the next sector (sigh, complex)
	mov bx,dx
	and bx,3	; get drive number
	add bx,bx
	add bx,geometry
	mov bx,[cs:bx]	; load geometry
	mov ax,cx
	inc cx		; start by hoping the best
	and al,0x3f
	cmp al,bl	; already was the max sector number?
	jb short havenext	; no? then we are done.
	dec cx		; IMPORTANT !
	and cl,0xc0
	inc cx		; wrap sector to 1
	cmp dh,bh	; already max head?
	pushf
	inc dh		; go to the next head
	popf
	jb short havenext	; if we had heads left, we are done
	mov dh,0	; wrap head to 0
	shr cl,6
	xchg cl,ch	; create proper 10bit cylinder number
	inc cx		; *** would be very strange to reach 0x400 here
			; *** reaching the max is also not checked here
	xchg cl,ch
	shl cl,6	; create warped "normal" cylinder number again
	inc cx		; sector is 1, again
havenext:
	pop ax
	pop bx
	ret

; -------------------------------------------------------------------

	; more helper functions

callold:	pushf		; calls the original int 0x13
		db 9ah          ; CALL DWORD ...
oldi13b:        dd 0f000a25fh   ; old int 0x13 vector again
		ret

meep:	pusha	; to have some feedback: BEEP and show AX and
		; message pointed to by cs:meepwhy, if any
	push ax	; save for later
	xor si,si
	xchg si,[cs:meepwhy]	; show message only once :-)
	or si,si
	jz short nomeeptext
meepmsgloop:
	mov bx,7
	mov ah,0x0e
	mov al,[cs:si]
	or al,al
	jz short nomeepmeep	; so only beep if no message
	inc si
		int 0x10	; show char
	jmp short meepmsgloop
nomeeptext:
        mov bx,7
        mov ax,0x0e07
	        int 0x10	; BEEP
nomeepmeep:
	pop ax
	push ax
	mov al,ah
		call showal	; show AH
	pop ax
		call showal	; show AL
	or si,si
	jz nomeepwait		; wait a while if we had a message
	mov bx,222		; *** adjust if you want
mpwt2:	mov cx,-1
mpwt1:	
	; in al,0x80		; I/O is slow, good...
	loop mpwt1		; worst case: 1/30 sec on a 386
	dec bx
	jnz short mpwt2
nomeepwait:
        popa
	ret

showal:			; helper func to show AL as hex
	push ax
	mov ah,0
	mov bx,ax
	shr bx,4
	mov al,[cs:hexa+bx]
	mov bx,7
	mov ah,0x0e
		int 0x10	; high hex digit
	pop ax
	and ax,0x000f
	mov bx,ax
	mov al,[cs:hexa+bx]
	mov bx,7
	mov ah,0x0e
		int 0x10	; low hex digit
	ret

hexa	db '0123456789abcdef'

; -------------------------------------------------------------------

	; status table handling functions

findbin:	; find bin matching location CX DX,
		; return carry or bin in AX (and update LRU)
	push si
	push cx
	push dx
	push bp
	mov bp,cx	; store cx for later use
	and dl,3	; only 2 lsb of drive number matter
	mov cx,[cs:sectors]
	mov si,table
fbloop:	mov ax,[cs:si]
	and al,3	; LRU does not matter now
	cmp ax,dx	; head and drive the same?
	jnz short fbmiss
	cmp bp,[cs:si+2]	; bp is old cx...
	jnz short fbmiss	; yes, but cylinder/sector differed
			; we hit the point :-)
	mov al,[cs:si]
	cmp al,0xfc	; LRU already at max?
	jae short fbmax
	add byte [cs:si],4	; increase importance / LRU by 1
fbmax:	sub si,table
	shr si,2	; each table entry is 4 byte, 32 bit
	mov ax,si	; this is the bin with our sector!
	clc		; indicate success
	jmp short fbout	; return the bin found (in AX)
fbmiss:	add si,4
	loop fbloop	; try the next entry
	mov ax,-1	; (should not be used anyway)
	stc		; indicate failure
fbout:	pop bp
	pop dx
	pop cx
	pop si
	ret


newbin:		; find a new free bin or reuse a bin
		; for location CX DX. update LRU, return bin in AX
		; > modified to be "round robin", so if all bins are
		; > equally important, we get "FIFOish" behaviour!
	push si
	push cx
	push bp
	push di			; >
	mov bp,cx	; store cx for entry in list
	mov ah,3	; start the search with the wish "free bin"
	mov di,[cs:sectors]	; >
	shl di,2		; > 32bit/4byte per table entry
	add di,table		; > pointer to first byte past table
nbnext:	mov cx,[cs:sectors]
	mov si,[cs:tabpos]	; > modified for "round robin"
nbloop:	mov al,[cs:si]	; 6 msb are LRU
	cmp al,ah	; LRU as we wish?
	jbe nbthis
	add si,4	; try NEXT bin to find a reusable one...
	cmp si,di		; >
	jb nbinrange		; > still inside table?
	mov si,table		; > wrap to beginning
nbinrange:
	loop nbloop
	add ah,4	; reduce "needed freeness for reusal" by 1 LRU unit
	cmp ah,0xfc+3
	jnz short nbnext	; *** search again if acceptable (slow?)
		call scale	; if all are max important, rescale importance
	mov si,table	; return 1st bin if everything else fails
nbthis:	mov ax,dx
	and al,3	; mask: only store 2 lsb of drive number
	add al,4	; set LRU value / importance to 1
	mov [cs:si],ax		; store LRU, drive, head in table
	mov [cs:si+2],bp	; store old cx (cylinder and sector) in table
	mov ax,si		; > update "round robin" search starting pos
	add ax,4		; > next 32bit/4byte entry... 
	cmp ax,di		; >
	jb nbinrange2		; >
	mov ax,table		; > wrap to beginning
nbinrange2:
	mov [cs:tabpos],ax	; > store new "round robin" starting point
	sub si,table
	shr si,2	; each table entry is 4 BYTE, 32 BIT !
	mov ax,si	; the bin!
	pop di			; >
	pop bp
	pop cx
	pop si
	ret


flush:	push di	; marks every table entry as empty
	push cx
	push eax
	mov cx,[cs:sectors]
	mov di,table
	mov eax,0xffffff00	; drive 0, unused, CHS=-1
flloop:	mov [cs:di],eax
	add di,4
	loop flloop
	pop eax
	pop cx
	pop di
	ret


scale:	mov word [cs:meepwhy],fullerr
		call meep	; complain about full cache
	push di	; marks every table entry as less important
	push cx
	push ax
	mov cx,[cs:sectors]
	mov di,table
scloop:	mov al,[cs:di]
	cmp al,4+3	; empty or least important already?
	jbe noscale
	sub al,4	; decrement importance
	mov [cs:di],al	; write new importance
noscale:
	add di,4
	loop scloop
	pop ax
	pop cx
	pop di
	ret

; -------------------------------------------------------------------

	; end of install code, needs to be resident (because of flush)

resinst:
		call flush	; mark cache as empty
	push eax
	xor ax,ax
	mov es,ax		; ES gets reset below anyhow
	mov eax,[es:0x13+0x13+0x13+0x13]	; read old vector
	mov [oldi13a],eax			; save it for chain
	mov [oldi13b],eax			; save if for call
	mov ax,cs
	shl eax,16				; new vector: segment
	mov ax,NEWI13				; new vector: offset
	mov [es:0x13+0x13+0x13+0x13],eax	; hook (no cli, atomic)
	pop eax
quitgood:
        les bx,[cs:pb]			; we need our driver thing again!
	mov word [es:bx+0x0e],ente	; stay resident
	mov ax,[cs:sectors]	; ... ds: prefix needed, see above
	shl ax,1
	shl ax,1
	add [es:bx+0x0e],ax	; table: 4 byte per cached sector
			; *** (overflow check not important for << 8 MBy)
	mov word [cs:meepwhy],meeperr
	mov ax,[es:bx+0x0e]
		call meep	; give size and test meep
	mov ax,0xe0d
	mov bx,7
		int 0x10	; CR
	mov ax,0xe0a
	mov bx,7
		int 0x10	; LF
quitpop:
	pop ds		; restore all regs and return
	popa
.8086
quitinst:
	mov byte [cs:running],1	; mark init as done
        jmp nix

; -------------------------------------------------------------------

fish2err:
	db 'HDDCACHE now off: dangerous call INT 0x13.',7,0
fisherr	db 'HDDCACHE flush: INT 0x13.',7,0
wrerr	db 'HDDCACHE flush: write error.',7,0
rderr	db 'HDDCACHE: read error.',7,0
xmserr	db 'HDDCACHE: XMS error.',7,0
fullerr	db 'HDDCACHE: cache full, scaling ',7,0
meeperr	db 'driver size: ',0
crlf	db 13,10,10

hello	db 'Hard disk XMS/386 read cache, E. Auer eric@coli.uni-sb.de 10/2001'
	db 13,10
	db 'License: GPL. Caching drives 0x80..0x83 using int 0x13.'
	db 13,10,0
align 4
	; the rest will be overwritten by our (aligned) status table
	; format: first byte is (DL && 3) || (importance << 2)
	; next byte is DH, then we have the CX word
table
ente	db 0	; first free byte marker, table size will be added

; -------------------------------------------------------------------

	; main install and setup routine follows
.8086
install:
;	cli	; cli not really needed...
	mov word [es:bx+0x10],cs	; pointer to first free byte
	mov word [es:bx+0x0e],NEWI13	; if load failed...

; test if we have 386 or better: 
	push ax
	pushf   ; save flags
		xor ax,ax
		push ax
		popf	; try to clear all bits
		pushf
	        pop ax
	and ax,0f000h
	cmp ax,0f000h
	jz noinst1	; 4 msb stuck to 1: 808x or 80186
		mov ax,0f000h
		push ax
		popf	; try to set 4 msb
		pushf
		pop ax
	test ax,0f000h
	jz noinst1	; 4 msb stuck to 0: 80286
	jmp short okinst1
noinst1:	; failed, no 386 found
	push bx
	push si
	push ds
	mov ax,cs
	mov ds,ax
	cld			; flags are saved :-)
	mov si,err386
		call strtty	; complain: no 386 found
	mov si,hello
		call strtty	; show banner
	pop ds
	pop si
	pop bx
	pop ax
	popf
	jmp quitinst
.386
okinst1:
	popf	; good, it is a 386, now restore flags
	pop ax

	pusha
	push ds
	mov ax,cs
	mov ds,ax		; save all regs and set DS to our CS
	cld			; flags are saved :-)

; test if we have XMS and allocate XMS, fail if none/not enough found

	mov ax,0x4300
		push ax	; for debugging: put "why" code on stack
	mov bl,-1		; clear errorcode value
		int 0x2f	; xms installation check
	cmp al,0x80
	jnz short noxms		; xms not present
	mov ax,0x4310
	push es
		int 0x2f	; get xms call vector (may clobber regs)
	mov [xmsvec],bx
	mov [xmsvec+2],es
	pop es
		pop ax	; for debugging: remove "why" code
	mov ah,8
		push ax	; for debugging: new "why" code
		call far [xmsvec]	; check amount of free XMS
	or ax,ax
	jz short noxms		; errorcode in bl if ax zero
	mov bx,[ds:sectors]	; ds: prefix needed, see above
	inc bx
	shr bx,1	; amount of XMS we need, kbyte / use 512 by sectors
	cmp dx,bx		; DX total, AX biggest chunk (kbyte)
	jb short noxms
	cmp ax,bx
		pop ax	; for debugging: remove "why" code
		push word 0xfe00	; for debugging: new "why" code
	jb short noxms		; not enough XMS free
		pop ax	; for debugging: remove "why" code
	mov ah,9
		push ax	; for debugging: new "why" code
	mov dx,bx
		call far [xmsvec]	; alloc DX kbyte for us
	or ax,ax
	jz short noxms		; errorcode in bl if ax zero
		pop ax	; for debugging: remove "why" code
	mov [xmshandle],dx	; our handle
	; (to free the memory, we would use function 0x0a and handle in DX)
	; (only other function used is copy, 0x0b)
	jmp short findgeom

noxms:	
		pop ax		; meep displays errorcode from bl...
		mov al,bl	; ...and the "why" code from stack.hi
	mov word [cs:meepwhy],xmserr
		call meep	; give feedback
	mov si,xmserr2		; no or not enough XMS found
		call strtty	; minimalistic error message
	mov si,hello
		call strtty	; show banner
	jmp quitpop

; now we have the XMS we need, after flushing the status table and
; finding the geometry we can hook int 0x13 and return...

findgeom:
	push es
	push cs
	pop es			; so we need no cs: prefix for stosb...
	mov di,geometry
	cld			; flags are saved :-)
	mov ah,08		; read geometry and number of hard disks
	mov dl,0x80		; HDA
	push es
	push di
		int 0x13	; find geometry (bl type, dl drives, dhcx geo)
	pop di			; (esdi -> drive param tab, only for floppies)
	pop es
	jc nodrive
	or dl,dl
	jz nodrive
	and cl,63
	mov al,cl
	stosb			; save max sector num
	mov al,dh
	stosb			; save max head num
	cmp dl,1
	jz short alldrives	; only one hard disk? then we are done
	mov ah,08
	mov dl,0x81		; HDB
	push es
	push di
		int 0x13	; find next geom, no further error checks!
	pop di
	pop es
	and cl,63
	mov al,cl
	stosb
	mov al,dh
	stosb
	cmp dl,2
	jz short alldrives	; yes, it is an unrolled loop...
	mov ah,08
	mov dl,0x82		; HDC
	push es
	push di
		int 0x13	; find next geom, no further error checks!
	pop di
	pop es
	and cl,63
	mov al,cl
	stosb
	mov al,dh
	stosb
	cmp dl,3
	jz short alldrives	; yes, it is an unrolled loop...
	mov ah,08
	mov dl,0x83		; HDD
	push es
	push di
		int 0x13	; find next geom, no further error checks!
	pop di
	pop es
	and cl,63
	mov al,cl
	stosb
	mov al,dh
	stosb
	jmp short alldrives	; if there are more drives, we ignore them!
	
nodrive:			; no drives to cache, not fatal
	mov si,errdrv
		call strtty	; very useful error message :-)
alldrives:			; done with all drives
	pop es
	mov si,hello		; the common HELLO MESSAGE:
		call strtty	; show banner


	jmp resinst	; jump out of the way, rest of inst will be
			; overwritten with the status table!

; -------------------------------------------------------------------


.8086
strtty: mov bx,7	; print out some text string using bios
        mov ah,0x0e
        lodsb
        or al,al
        jz short sttend
	        int 0x10
        jmp short strtty
sttend: ret

xmserr2	db 'This software needs a 386 and enough free XMS memory'
	db 13,10,0

err386	db 'This software needs at least an 80386 CPU'
	db 13,10,0

errdrv	db 'Zero hard drives installed, will only pretend caching'
	db 13,10,0

; dat     ENDS
; END
